// ======================================================================
//
// gl.h
// jeff grills
//
// copyright 1998 origin systems inc
//
// ======================================================================

#ifndef GL_H
#define GL_H

// ======================================================================
// includes

struct GL_input;
struct GL_output;

#include "Gl.def"
#include "gl_abstract.h"
#include "Material.h"
#include "Palette.h"

// ======================================================================

class Gl
{
	friend class MoviePlayer;

private:

	// prevent objects of this class from being created
	Gl(void);
	Gl(const Gl &);
	Gl &operator =(const Gl &);

public:

	enum
	{
		RasterMajorWu   = 1,
		RasterMajorD3D  = 3,
		RasterMajor3dfx = 5,
		RasterMajorWu2  = 6
	};

private:

	enum
	{
		CLIP_INFO_STACK = 4
	};

	typedef GL_abstract *(FAR *CreateGl)(void);
	typedef void         (FAR *DestroyGl)(GL_abstract *);

private:

	// dll stuff
	static HINSTANCE      dllHandle;
	static int            dllNumber;
	static CreateGl       createGl;
	static DestroyGl      destroyGl;
	static int            rasterMajor;
	static int            rasterMinor;
	static GL_abstract   *gl;

	// screen size
	static int            screenWidth;
	static int            screenHeight;
	static int            screenDepthBytes;

	// clipping stuff
	static ClipInfo      *clipInfo;
	static ClipInfo       clipInfoStack[CLIP_INFO_STACK];

	static int            materialToken;

	// this is here for the movieplayer to peek at only
	static byte           glRMovieSizeShift;
	static byte           glRMoviePlaceShift;
	static byte           glGMovieSizeShift;
	static byte           glGMoviePlaceShift;
	static byte           glBMovieSizeShift;
	static byte           glBMoviePlaceShift;

	static int                   alternatingBackgroundMaterial;
	static const StaticMaterial *backgroundMaterial;
	static const StaticMaterial *currentBackgroundMaterial;
	static const StaticMaterial *singleFrameBackgroundMaterial;

	static int            frameSkip;

	static int            textureMemory;

	static bool           useSuperDetailFlag;
	static bool           lightingRGB;
	static bool           translucency;
	static bool           fog;
	static bool           bilinearFiltering;
	static bool           lensFlares;
	
	static bool           warningOneMegVideoCard;
	static bool           warningBackBufferInSystemRam;
	static bool           warningBankSwitched;

	static bool           mode2d;

	// cache variables needed to reset the Gl state when it gets reloaded after task switching
	static PrimitiveProcessing   primitiveProcessing;
	static const StaticMaterial *material;
	static bool                  alphaBlendEnabled;
	static AlphaBlendFunction    alphaBlendSource;
	static AlphaBlendFunction    alphaBlendDestination;

private:

	static void   skipFrame(int count=1);
	static void   loadDll(int newNumber);
	static void   unloadDll(void);

	static void   modeError(void);
	static void   verifyMode2d(void);
	static void   verifyMode3d(void);

public:

	// installation and removal for a static class
	static void   install(void);
	static void   remove(void);

	// rasterizer DLL functions
	static void   setRasterizer(int major, int minor);
	static int    getRasterMajor(void);
	static int    getRasterMinor(void);

	// debugging stuff
	static void   debugDump(void);

	// resize the screen
	static void   resize(int newWidth, int newHeight, int newDepthBytes, bool force=false);

	// size functions
	static int    width(void);
	static int    height(void);
	static int    depthBytes(void);
	static int    depthBits(void);

	// clipping functions
	static void            clipPush(int x0, int y0, int x1, int y1);
	static void            clipOff(void);
	static void            clipPop(void);
	static const ClipInfo &getClipInfo(void);

	// rasterizer capability querying
	static bool useSuperDetail(void);
	static bool isLightingRGB(void);
	static bool hasTranslucency(void);
	static bool hasFog(void);
	static bool hasBilinearFiltering(void);
	static bool hasLensFlares(void);
	static int  getTextureMemory(void);
	static bool oneMegVideoCardWarning(void);
	static bool backBufferInSystemRamWarning(void);
	static bool bankSwitchedWarning(void);

	// set the output window for clipping
	static void setDrawWindow(HWND newDrawWindow);

	// handle application activate and inactivate
	static void setActive(bool active);

	// access functions
	static void lock(void);
	static void unlock(void);
	static void screenShot(const char *fileName);
	static bool gdiVisible(void);

	// set processing of primitives
	static void setAlphaBlendEnabled(bool enabled);
	static void setAlphaBlendFunctions(AlphaBlendFunction source=ABF_One, AlphaBlendFunction destination=ABF_Zero);
	static void setPrimitiveProcessing(PrimitiveProcessing primitiveProcessing);

	// palette functions
	static PaletteData *paletteCreateData(const Palette *pal);
	static void         paletteDestroyData(PaletteData *palData);

	// material data functions
	static MaterialData *materialCreateData(const StaticMaterial *material);
	static void          materialDestroyData(MaterialData *mat);
	static void          materialDestroyed(const Material *material);

	// material functions
	static void setMaterial(const StaticMaterial *material);
	static void setBackgroundMaterial(const StaticMaterial *material);
	static void setSingleFrameBackgroundMaterial(const StaticMaterial *material);

	// begin and end camera renders
	static void beginRender(const Camera &camera);
	static void endRender(void);

	// 3d transformation functions
	static void setLightingEnabled(bool lightingEnabled);
	static void setLightList(Light **lightList, int lightCount);
	static void setObjectToWorld(const Xform &objectToWorld);
	static bool setVertexList(const Vector *vertex, int vertexCount, const Vector *normal, int normalCount);
	
	// convenient high-level drawing primitives
	static void point(int x, int y, real z=CONST_REAL(1), real ooz=CONST_REAL(1));
	static void line(int x0, int y0, int x1, int y1, real z=CONST_REAL(1), real ooz=CONST_REAL(1));
	static void rect(int x0, int y0, int x1, int y1, bool filled=false, real z=CONST_REAL(1), real ooz=CONST_REAL(1));
	static void point(const Vector &v);
	static void line(const Vector &v0, const Vector &v1);

	// object-space drawing primitives
	static void drawPointList(const UVertex *vertex, int numberOfVertices);
	static void drawLineList(const UVertex *vertex, int numberOfVertices);
	static void drawLineStrip(const UVertex *vertex, int numberOfVertices);
	static void drawTriangleList(const UVertex *vertex, int numberOfVertices);
	static void drawTriangleStrip(const UVertex *vertex, int numberOfVertices);
	static void drawTriangleFan(const UVertex *vertex, int numberOfVertices);

	// screen-space drawing primitives
	static void drawPointList(const TVertex *vertex, int numberOfVertices);
	static void drawLineList(const TVertex *vertex, int numberOfVertices);
	static void drawLineStrip(const TVertex *vertex, int numberOfVertices);
	static void drawTriangleList(const TVertex *vertex, int numberOfVertices);
	static void drawTriangleStrip(const TVertex *vertex, int numberOfVertices);
	static void drawTriangleFan(const TVertex *vertex, int numberOfVertices);

	// per-rasterizer special effects
	static bool specialEffect(int id, void *data);

	// movie function
	static void  movie(const uv_instance *uv);
};

// ======================================================================

inline int Gl::getRasterMajor(void)
{
	return rasterMajor;
}

// ----------------------------------------------------------------------

inline int Gl::getRasterMinor(void)
{
	return rasterMinor;
}

// ----------------------------------------------------------------------

inline int Gl::width(void)
{
	return screenWidth;
}

// ----------------------------------------------------------------------

inline int Gl::height(void)
{
	return screenHeight;
}

// ----------------------------------------------------------------------

inline int Gl::depthBytes(void)
{
	return screenDepthBytes;
}

// ----------------------------------------------------------------------

inline int Gl::depthBits(void)
{
	return (screenDepthBytes << 3);
}

// ----------------------------------------------------------------------

inline void Gl::setBackgroundMaterial(const StaticMaterial *material)
{
	backgroundMaterial = material;
}

// ----------------------------------------------------------------------

inline void Gl::setSingleFrameBackgroundMaterial(const StaticMaterial *material)
{
	singleFrameBackgroundMaterial = material;
}

// ----------------------------------------------------------------------

inline const ClipInfo &Gl::getClipInfo(void)
{
	return *clipInfo;
}

// ----------------------------------------------------------------------

inline void Gl::skipFrame(int count)
{
	frameSkip += count;
}

// ----------------------------------------------------------------------

inline void Gl::setDrawWindow(HWND newDrawWindow)
{
	gl->setDrawWindow(newDrawWindow);
}

// ----------------------------------------------------------------------

inline void Gl::setActive(bool active)
{
	if (gl)
		gl->setActive(active);
}

// ----------------------------------------------------------------------

inline void Gl::debugDump(void)
{
	gl->debugDump();
}

// ----------------------------------------------------------------------

inline void Gl::screenShot(const char *fileName)
{
	gl->screenShot(fileName);
}

// ----------------------------------------------------------------------

inline void Gl::setAlphaBlendEnabled(bool enabled)
{
	alphaBlendEnabled = enabled;
	gl->setAlphaBlendEnabled(enabled);
}

// ----------------------------------------------------------------------

inline void Gl::setAlphaBlendFunctions(AlphaBlendFunction source, AlphaBlendFunction destination)
{
	alphaBlendSource = source;
	alphaBlendDestination = destination;
	gl->setAlphaBlendFunctions(source, destination);
}

// ----------------------------------------------------------------------

inline void Gl::setPrimitiveProcessing(PrimitiveProcessing primitiveProcessing)
{
	gl->setPrimitiveProcessing(primitiveProcessing);
}

// ----------------------------------------------------------------------

inline void Gl::verifyMode2d(void)
{
	if (!mode2d)
		modeError();
}

// ----------------------------------------------------------------------

inline void Gl::verifyMode3d(void)
{
	if (mode2d)
		modeError();
}

// ----------------------------------------------------------------------

inline bool Gl::gdiVisible(void)
{
	return gl->gdiVisible();
}

// ----------------------------------------------------------------------

inline PaletteData *Gl::paletteCreateData(const Palette *pal)
{
	return gl->paletteCreateData(pal);
}

// ----------------------------------------------------------------------

inline void Gl::paletteDestroyData(PaletteData *palData)
{
	gl->paletteDestroyData(palData);
}

// ----------------------------------------------------------------------

inline void Gl::movie(const uv_instance *uv)
{
	materialToken = 0;
	verifyMode2d();
	gl->movie(uv);
}

// ----------------------------------------------------------------------

inline bool Gl::useSuperDetail(void)
{
	return useSuperDetailFlag;
}

// ----------------------------------------------------------------------

inline bool Gl::isLightingRGB(void)
{
	return lightingRGB;
}

// ----------------------------------------------------------------------

inline bool Gl::hasTranslucency(void)
{
	return translucency;
}

// ----------------------------------------------------------------------

inline bool Gl::hasFog(void)
{
	return fog;
}

// ----------------------------------------------------------------------

inline bool Gl::hasBilinearFiltering(void)
{
	return bilinearFiltering;
}

// ----------------------------------------------------------------------

inline bool Gl::hasLensFlares(void)
{
	return lensFlares;
}

// ----------------------------------------------------------------------

inline bool Gl::oneMegVideoCardWarning(void)
{
	return warningOneMegVideoCard;
}

// ----------------------------------------------------------------------

inline bool Gl::backBufferInSystemRamWarning(void)
{
	return warningBackBufferInSystemRam;
}

// ----------------------------------------------------------------------

inline bool Gl::bankSwitchedWarning(void)
{
	return warningBankSwitched;
}

// ----------------------------------------------------------------------

inline int Gl::getTextureMemory(void)
{
	return textureMemory;
}

// ----------------------------------------------------------------------

inline MaterialData *Gl::materialCreateData(const StaticMaterial *material)
{
	return gl->materialCreateData(material);
}

// ----------------------------------------------------------------------

inline void Gl::materialDestroyData(MaterialData *mat)
{
	gl->materialDestroyData(mat);
}

// ----------------------------------------------------------------------

inline void Gl::beginRender(const Camera &camera)
{
	verifyMode2d();
	gl->beginRender(camera);
	mode2d = false;
}

// ----------------------------------------------------------------------

inline void Gl::endRender(void)
{
	verifyMode3d();
	gl->endRender();
	mode2d = true;
}

// ----------------------------------------------------------------------

inline void Gl::setObjectToWorld(const Xform &objectToWorld)
{
	verifyMode3d();
	gl->setObjectToWorld(objectToWorld);
}

// ----------------------------------------------------------------------

inline void Gl::setLightingEnabled(bool lightingEnabled)
{
	gl->setLightingEnabled(lightingEnabled);
}

// ----------------------------------------------------------------------

inline void Gl::setLightList(Light **lightList, int lightCount)
{
	verifyMode3d();
	gl->setLightList(lightList, lightCount);
}

// ----------------------------------------------------------------------

inline void Gl::setMaterial(const StaticMaterial *mat)
{
	material = mat;

	if (!material)
	{
		if (materialToken)
		{
			materialToken = 0;
			gl->setMaterial(NULL);
		}
	}
	else
		if (material->getToken() != materialToken)
		{
			materialToken = material->getToken();
			gl->setMaterial(material);
		}
}

// ----------------------------------------------------------------------

inline void Gl::materialDestroyed(const Material *mat)
{
	if (mat && mat->getToken() == materialToken)
		setMaterial(NULL);
}

// ----------------------------------------------------------------------

inline bool Gl::setVertexList(const Vector *vertex, int vertexCount, const Vector *normal, int normalCount)
{
	verifyMode3d();
	return gl->setVertexList(vertex, vertexCount, normal, normalCount);
}

// ----------------------------------------------------------------------

inline void Gl::drawPointList(const UVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(materialToken == 0, ("Gl::drawPointList no material"));
	verifyMode3d();
	gl->drawPointList(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

inline void Gl::drawLineList(const UVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(materialToken == 0, ("Gl::drawLineList no material"));
	verifyMode3d();
	gl->drawLineList(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

inline void Gl::drawLineStrip(const UVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(materialToken == 0, ("Gl::drawLineStrip no material"));
	verifyMode3d();
	gl->drawLineStrip(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

inline void Gl::drawTriangleList(const UVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(materialToken == 0, ("Gl::drawTriangleList no material"));
	verifyMode3d();
	gl->drawTriangleList(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

inline void Gl::drawTriangleStrip(const UVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(materialToken == 0, ("Gl::drawTriangleStrip no material"));
	verifyMode3d();
	gl->drawTriangleStrip(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

inline void Gl::drawTriangleFan(const UVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(materialToken == 0, ("Gl::drawTriangleFan no material"));
	verifyMode3d();
	gl->drawTriangleFan(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

inline void Gl::drawPointList(const TVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(materialToken == 0, ("Gl::drawPointList no material"));
	verifyMode2d();
	gl->drawPointList(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

inline void Gl::drawLineList(const TVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(materialToken == 0, ("Gl::drawLineList no material"));
	verifyMode2d();
	gl->drawLineList(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

inline void Gl::drawLineStrip(const TVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(materialToken == 0, ("Gl::drawLineStrip no material"));
	verifyMode2d();
	gl->drawLineStrip(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

inline void Gl::drawTriangleList(const TVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(materialToken == 0, ("Gl::drawTriangleList no material"));
	verifyMode2d();
	gl->drawTriangleList(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

inline void Gl::drawTriangleStrip(const TVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(materialToken == 0, ("Gl::drawTriangleStrip no material"));
	verifyMode2d();
	gl->drawTriangleStrip(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

inline void Gl::drawTriangleFan(const TVertex *vertexList, int numberOfVertices)
{
	DEBUG_FATAL(materialToken == 0, ("Gl::drawTriangleFan no material"));
	verifyMode2d();
	gl->drawTriangleFan(vertexList, numberOfVertices);
}

// ----------------------------------------------------------------------

inline bool Gl::specialEffect(int id, void *data)
{
	materialToken = 0;
	return gl->specialEffect(id, data);
}

// ======================================================================

#endif
